#pragma once

// Genode includes
#include <base/heap.h>
#include <libc/component.h>
#include <util/fifo.h>

// gapfruit includes
#include <util/hexdump.h>

// local includes
#include "command.h"
#include "exception.h"
#include "socket_wrapper.h"


namespace Hsm {
	class Payload;
}

class Hsm::Payload : public Genode::Fifo<Payload>::Element
{
	private:
		Data_type _data_type = Data_type::dataTypeUnknown;
		const uint8_t* _payload_buffer = nullptr;
		uint32_t _payload_length = 0;
		Genode::Heap& _heap;

		uint32_t _padding_length() const
		{
			const uint32_t remainder = _payload_length % sizeof(uint32_t);
			return (remainder == 0) ? 0 : (sizeof(uint32_t) - remainder);
		}

		uint32_t _padded_length() const
		{
			return _payload_length + _padding_length();
		}

	public:
		Data_type data_type() { return _data_type; }
		const uint8_t* payload_buffer() { return _payload_buffer; }
		uint32_t payload_length() { return _payload_length; }

		Payload(const Hsm::Payload&) = delete;
		Payload operator=(const Hsm::Payload&) = delete;

		Payload(Genode::Heap& heap,
		        const Data_type data_type,
		        const uint8_t* payload_buffer,
		        const uint32_t payload_length)
			: _data_type(data_type), _payload_buffer(payload_buffer),
			  _payload_length(payload_length), _heap(heap)
		{
		}

		Payload(Genode::Heap& heap,
		        const uint8_t* message_buffer,
		        const uint32_t message_length)
			: _heap(heap)
		{
			if (message_length < (sizeof(Payload_header) - sizeof(uint32_t))) {
				Genode::error("Buffer too short to contain header");
				throw Buffer_too_short_exception();
			}

			const Payload_header* header = reinterpret_cast<const Payload_header*>(message_buffer);
			_data_type = static_cast<Data_type>(header->type);
			const uint8_t* current_buffer = static_cast<const uint8_t*>(message_buffer);

			if (header->length == 0xFFFF) {
				if (message_length < sizeof(Payload_header)) {
					Genode::error("Buffer too short to contain extended header");
					throw Buffer_too_short_exception();
				}
				_payload_length = header->long_length;
				if (message_length < sizeof(Payload_header) + _payload_length) {
					Genode::error("Buffer too short to contain extended header and payload");
					throw Buffer_too_short_exception();
				}
				current_buffer += sizeof(Payload_header);
			} else {
				_payload_length = header->length;
				if (message_length < sizeof(Payload_header) - sizeof(uint32_t) + _payload_length) {
					Genode::error("Buffer too short to contain short header and payload");
					throw Buffer_too_short_exception();
				}
				current_buffer += sizeof(Payload_header) - sizeof(uint32_t);
			}

			_payload_buffer = current_buffer;
		}

		uint32_t length() const
		{
			if (_padded_length() < 0xFFFF) {
				return _padded_length() + sizeof(Payload_header) - sizeof(uint32_t);
			} else {
				return _padded_length() + sizeof(Payload_header);
			}
		}

		void write_to_socket(const Socket_wrapper& socket) const
		{
			Payload_header header;
			Genode::memset(&header, 0, sizeof(header));
			header.type = _data_type;

			if (_payload_length < 0xFFFF) {
				header.length = static_cast<uint16_t>(_payload_length);
				socket.write(&header, sizeof(header) - sizeof(uint32_t));
			} else {
				header.length = 0xFFFF;
				header.long_length = _payload_length;
				socket.write(&header, sizeof(header));
			}

			socket.write(_payload_buffer, _payload_length);

			uint8_t padding[sizeof(uint32_t)];
			Genode::memset(padding, 0, sizeof(uint32_t));
			socket.write(padding, _padding_length());
		}
};

